import datetime
import enum
import math
import os
import random
import time
import typing
import uuid

import flask
import jinja2.filters
import tqdm
import ygojson

import yjviewer.search as search

from .locales import LOCALE_TRANSLATED
from .version import __version__

if os.path.exists(ygojson.AGGREGATE_DIR):
    ygodb = ygojson.load_from_file(aggregates_dir=ygojson.AGGREGATE_DIR)
else:
    ygodb = ygojson.load_from_internet(aggregates_dir=ygojson.AGGREGATE_DIR)
ygodb.regenerate_backlinks()

portenial_cotd = [x for x in ygodb.cards if x.images and x.images[0].card_art]
cards_of_the_day = [random.choice(portenial_cotd) for i in range(5)]
del portenial_cotd

app = flask.Flask(__name__)

ENUM_TRANSLATED: typing.Dict[enum.Enum, str] = {
    ygojson.CardType.MONSTER: "Monster",
    ygojson.CardType.SPELL: "Spell",
    ygojson.CardType.TRAP: "Trap",
    ygojson.CardType.TOKEN: "Token",
    ygojson.CardType.SKILL: "Skill",
    ygojson.Attribute.DARK: "DARK",
    ygojson.Attribute.DIVINE: "DIVINE",
    ygojson.Attribute.EARTH: "EARTH",
    ygojson.Attribute.FIRE: "FIRE",
    ygojson.Attribute.LIGHT: "LIGHT",
    ygojson.Attribute.WATER: "WATER",
    ygojson.Attribute.WIND: "WIND",
    ygojson.MonsterCardType.FUSION: "Fusion",
    ygojson.MonsterCardType.LINK: "Link",
    ygojson.MonsterCardType.PENDULUM: "Pendulum",
    ygojson.MonsterCardType.RITUAL: "Ritual",
    ygojson.MonsterCardType.SYNCHRO: "Synchro",
    ygojson.MonsterCardType.XYZ: "Xyz",
    ygojson.Race.BEASTWARRIOR: "Beast-Warrior",
    ygojson.Race.ZOMBIE: "Zombie",
    ygojson.Race.FIEND: "Fiend",
    ygojson.Race.DINOSAUR: "Dinosaur",
    ygojson.Race.DRAGON: "Dragon",
    ygojson.Race.BEAST: "Beast",
    ygojson.Race.ILLUSION: "Illusion",
    ygojson.Race.INSECT: "Insect",
    ygojson.Race.WINGEDBEAST: "Winged-Beast",
    ygojson.Race.WARRIOR: "Warrior",
    ygojson.Race.SEASERPENT: "Sea-Serpent",
    ygojson.Race.AQUA: "Aqua",
    ygojson.Race.PYRO: "Pyro",
    ygojson.Race.THUNDER: "Thunder",
    ygojson.Race.SPELLCASTER: "Spellcaster",
    ygojson.Race.PLANT: "Plant",
    ygojson.Race.ROCK: "Rock",
    ygojson.Race.REPTILE: "Reptile",
    ygojson.Race.FAIRY: "Fairy",
    ygojson.Race.FISH: "Fish",
    ygojson.Race.MACHINE: "Machine",
    ygojson.Race.DIVINEBEAST: "Divine-Beast",
    ygojson.Race.PSYCHIC: "Psychic",
    ygojson.Race.CREATORGOD: "Creator-God",
    ygojson.Race.WYRM: "Wyrm",
    ygojson.Race.CYBERSE: "Cyberse",
    ygojson.Classification.EFFECT: "Effect",
    ygojson.Classification.TUNER: "Tuner",
    ygojson.Classification.SPECIALSUMMON: "Special Summon",
    ygojson.Classification.NORMAL: "Normal",
    ygojson.Classification.PENDULUM: "Pendulum",
    ygojson.Ability.FLIP: "Flip",
    ygojson.Ability.GEMINI: "Gemini",
    ygojson.Ability.SPIRIT: "Spirit",
    ygojson.Ability.TOON: "Toon",
    ygojson.Ability.UNION: "Union",
    ygojson.LinkArrow.BOTTOMCENTER: "⬇",
    ygojson.LinkArrow.BOTTOMLEFT: "↙",
    ygojson.LinkArrow.BOTTOMRIGHT: "↘",
    ygojson.LinkArrow.MIDDLELEFT: "⬅",
    ygojson.LinkArrow.MIDDLERIGHT: "➡",
    ygojson.LinkArrow.TOPCENTER: "⬆",
    ygojson.LinkArrow.TOPLEFT: "↖",
    ygojson.LinkArrow.TOPRIGHT: "↗",
    ygojson.SubCategory.CONTINUOUS: "Continuous",
    ygojson.SubCategory.COUNTER: "Counter",
    ygojson.SubCategory.EQUIP: "Equip",
    ygojson.SubCategory.FIELD: "Field",
    ygojson.SubCategory.NORMAL: "Normal",
    ygojson.SubCategory.QUICKPLAY: "Quick-Play",
    ygojson.SubCategory.RITUAL: "Ritual",
    ygojson.Legality.UNLIMITED: "Unlimited",
    ygojson.Legality.FORBIDDEN: "Forbidden",
    ygojson.Legality.LIMIT1: "Limit 1",
    ygojson.Legality.LIMIT2: "Limit 2",
    ygojson.Legality.LIMIT3: "Limit 3",
    ygojson.Legality.LIMITED: "Limited",
    ygojson.Legality.SEMILIMITED: "Semilimited",
    ygojson.Legality.UNRELEASED: "Unreleased",
    ygojson.VideoGameRaity.NORMAL: "N",
    ygojson.VideoGameRaity.RARE: "R",
    ygojson.VideoGameRaity.SUPER: "SR",
    ygojson.VideoGameRaity.ULTRA: "UR",
    ygojson.SetEdition.FIRST: "1st Edition",
    ygojson.SetEdition.UNLIMTED: "Unlimited",
    ygojson.SetEdition.LIMITED: "Limited Edition",
    ygojson.SetEdition.NONE: "N/A",
    ygojson.Format.DUELLINKS: "Duel Links",
    ygojson.Format.MASTERDUEL: "Master Duel",
    ygojson.Format.OCG: "OCG",
    ygojson.Format.TCG: "TCG",
    ygojson.Format.SPEED: "Speed Duel",
    ygojson.CardRarity.COMMON: "C",
    ygojson.CardRarity.SHORTPRINT: "SP/SSP",
    ygojson.CardRarity.RARE: "R",
    ygojson.CardRarity.SUPER: "SR",
    ygojson.CardRarity.ULTRA: "UR",
    ygojson.CardRarity.ULTIMATE: "UtR",
    ygojson.CardRarity.SECRET: "ScR",
    ygojson.CardRarity.ULTRASECRET: "UScR",
    ygojson.CardRarity.PRISMATICSECRET: "PScR",
    ygojson.CardRarity.GHOST: "GR/HR",
    ygojson.CardRarity.PARALLEL: "PR",
    ygojson.CardRarity.COMMONPARALLEL: "PC/NPR",
    ygojson.CardRarity.RAREPARALLEL: "RPR",
    ygojson.CardRarity.SUPERPARALLEL: "SPR",
    ygojson.CardRarity.ULTRAPARALLEL: "UPR",
    ygojson.CardRarity.DTPC: "DPC/DNPR",
    ygojson.CardRarity.DTPSP: "DPSP/DNRPR",
    ygojson.CardRarity.DTRPR: "DRPR",
    ygojson.CardRarity.DTSPR: "DSPR",
    ygojson.CardRarity.DTUPR: "DUPR",
    ygojson.CardRarity.DTSCPR: "DScPR",
    ygojson.CardRarity.GOLD: "Gold",
    ygojson.CardRarity.TENTHOUSANDSECRET: "10,000ScR",
    ygojson.CardRarity.TWENTITHSECRET: "20ScR",
    ygojson.CardRarity.COLLECTORS: "CR",
    ygojson.CardRarity.EXTRASECRET: "EScR",
    ygojson.CardRarity.EXTRASECRETPARALLEL: "EScPR",
    ygojson.CardRarity.GOLDGHOST: "G/GR",
    ygojson.CardRarity.GOLDSECRET: "GScR",
    ygojson.CardRarity.STARFOIL: "Starfoil",
    ygojson.CardRarity.MOSAIC: "Mosaic",
    ygojson.CardRarity.SHATTERFOIL: "Shatterfoil",
    ygojson.CardRarity.GHOSTPARALLEL: "GPR",
    ygojson.CardRarity.PLATINUM: "PtR",
    ygojson.CardRarity.PLATINUMSECRET: "PtScR",
    ygojson.CardRarity.PREMIUMGOLD: "PGR",
    ygojson.CardRarity.TWENTYFIFTHSECRET: "25ScR",
    ygojson.CardRarity.SECRETPARALLEL: "PScR",
    ygojson.CardRarity.STARLIGHT: "StR/AltR",
    ygojson.CardRarity.PHARAOHS: "PUR",
    ygojson.CardRarity.KCCOMMON: "KCC",
    ygojson.CardRarity.KCRARE: "KCR",
    ygojson.CardRarity.KCSUPER: "KCSR",
    ygojson.CardRarity.KCULTRA: "KCUR",
    ygojson.CardRarity.KCSECRET: "KCScR",
    ygojson.CardRarity.MILLENIUM: "MR",
    ygojson.CardRarity.MILLENIUMSUPER: "MSR",
    ygojson.CardRarity.MILLENIUMULTRA: "MUR",
    ygojson.CardRarity.MILLENIUMSECRET: "MScR",
    ygojson.CardRarity.MILLENIUMGOLD: "MGR",
}

FORMAT_TRANSLATED = {
    "tcg": "TCG",
    "speed": "Speed Duel",
    "ocg": "OCG",
    "ocg-kr": "Korean OCG",
    "ocg-sc": "Chinese OCG",
    "masterduel": "Master Duel",
    "duellinks": "Duel Links",
}

FORMAT_TO_LOCALES = {
    "tcg": ["en", "na", "eu", "au", "oc", "es", "sp", "pt", "de", "it", "fr", "fc"],
    "speed": ["en", "na", "eu", "au", "oc", "es", "pt", "de", "it", "fr", "fc"],
    "ocg": ["jp", "ja", "ae", "tc", "zh-TW"],
    "ocg-kr": ["kr", "ko"],
    "ocg-sc": ["sc", "zh-CN"],
}

LOCALE_TO_FORMAT = {
    "en": ygojson.Format.TCG,
    "na": ygojson.Format.TCG,
    "eu": ygojson.Format.TCG,
    "au": ygojson.Format.TCG,
    "oc": ygojson.Format.TCG,
    "es": ygojson.Format.TCG,
    "sp": ygojson.Format.TCG,
    "pt": ygojson.Format.TCG,
    "de": ygojson.Format.TCG,
    "it": ygojson.Format.TCG,
    "fr": ygojson.Format.TCG,
    "fc": ygojson.Format.TCG,
    "jp": ygojson.Format.OCG,
    "ja": ygojson.Format.OCG,
    "ae": ygojson.Format.OCG,
    "tc": ygojson.Format.OCG,
    "zh-TW": ygojson.Format.OCG,
    "kr": ygojson.Format.OCG,
    "ko": ygojson.Format.OCG,
    "sc": ygojson.Format.OCG,
    "zh-CN": ygojson.Format.OCG,
}


@app.template_filter()
def zfill(s, n: int):
    return str(s).zfill(n)


@app.template_filter()
def translateenum(e: enum.Enum):
    return ENUM_TRANSLATED.get(e, e)


@app.template_filter()
def translateenums(es: typing.Iterable[enum.Enum]) -> typing.Iterable:
    for e in es:
        yield translateenum(e)


@app.template_filter()
def translateformat(f: str) -> str:
    return FORMAT_TRANSLATED.get(f, f)


@app.template_filter()
def translatelocale(l: str) -> str:
    return LOCALE_TRANSLATED.get(l, l)


@app.template_filter()
def currentlegality(
    card: ygojson.Card, format: str
) -> typing.Optional[ygojson.Legality]:
    if card.illegal:
        return ygojson.Legality.FORBIDDEN

    if format in card.legality:
        legality = card.legality[format]
        if legality.current:
            return legality.current
        if legality.history:
            return legality.history[-1].legality

    locales = FORMAT_TO_LOCALES.get(format, [])
    dates = []
    for set_ in card.sets:
        for localecode in locales:
            if localecode in set_.locales:
                locale = set_.locales[localecode]
                if locale.date:
                    dates.append(locale.date)
        else:
            if set_.date:
                dates.append(set_.date)

    if not dates:
        return None
    today = datetime.datetime.now().date()
    if all(date > today for date in dates):
        return ygojson.Legality.UNRELEASED
    else:
        return ygojson.Legality.UNLIMITED


@app.template_filter()
def getsetlocales(
    set_: ygojson.Set,
) -> typing.Iterable[typing.Optional[ygojson.SetLocale]]:
    if not set_.locales:
        yield None
        return
    yield from set_.locales.values()


@app.template_filter()
def getsetlocalecontents(
    set_: ygojson.Set, locale: typing.Optional[ygojson.SetLocale]
) -> typing.Iterable[ygojson.SetContents]:
    if not locale:
        yield from set_.contents
        return
    for content in set_.contents:
        if locale in content.locales:
            yield content


@app.template_filter()
def getseteditions(
    set_: ygojson.Set, locale: typing.Optional[ygojson.SetLocale]
) -> typing.Iterable[typing.Optional[ygojson.SetEdition]]:
    result = set()
    if not locale:
        for content in set_.contents:
            result.update(content.editions)
    else:
        result.update(locale.editions)
    if not result:
        yield None
    else:
        yield from result


CARD_BACK_URL = "https://ms.yugipedia.com//e/e5/Back-EN.png"


@app.template_filter()
def printingimage(
    set_: ygojson.Set,
    card: ygojson.Card,
    printing: ygojson.CardPrinting,
    locale: typing.Optional[ygojson.SetLocale],
    edition: typing.Optional[ygojson.SetEdition],
) -> str:
    if not edition:
        edition = ygojson.SetEdition.NONE

    if not printing or not locale:
        if card.images:
            return card.images[0].card_art or CARD_BACK_URL
        return CARD_BACK_URL

    printing = ygodb.printings_by_id[
        printing.id
    ]  # Flask loves to make copies of random objects

    if edition in locale.card_images and printing in locale.card_images[edition]:
        return locale.card_images[edition][printing]
    if edition == ygojson.SetEdition.NONE:
        for e in locale.card_images:
            if printing in locale.card_images[e]:
                return locale.card_images[e][printing]
    if (
        ygojson.SetEdition.NONE in locale.card_images
        and printing in locale.card_images[ygojson.SetEdition.NONE]
    ):
        return locale.card_images[ygojson.SetEdition.NONE][printing]
    return CARD_BACK_URL


@app.template_filter()
def dbsetlinks(db_ids: typing.Iterable[int]) -> typing.Iterable[str]:
    return [
        f'<a href="https://www.db.yugioh-card.com/yugiohdb/card_search.action?ope=1&pid={id}&rp=99999&request_locale=en">{id}</a>'
        for id in db_ids
    ]


@app.template_filter()
def setgenericboximage(set_: ygojson.Set) -> str:
    for preferred_locale in ["en", "na", "ja", "jp"]:
        if (
            preferred_locale in set_.locales
            and set_.locales[preferred_locale].box_image
        ):
            return set_.locales[preferred_locale].box_image or ""
    for locale in set_.locales.values():
        if locale.box_image:
            return locale.box_image
    for content in set_.contents:
        if content.box_image:
            return content.box_image
    return ""


@app.template_filter()
def setgenericpackimage(set_: ygojson.Set) -> str:
    for preferred_locale in ["en", "na", "ja", "jp"]:
        if preferred_locale in set_.locales and set_.locales[preferred_locale].image:
            return set_.locales[preferred_locale].image or ""
    for locale in set_.locales.values():
        if locale.image:
            return locale.image
    for content in set_.contents:
        if content.image:
            return content.image
    return ""


@app.template_filter()
def setgenericimage(set_: ygojson.Set) -> str:
    return setgenericpackimage(set_) or setgenericboximage(set_) or CARD_BACK_URL


@app.template_filter()
def seriesgenericimage(series: ygojson.Series) -> str:
    return CARD_BACK_URL


@app.template_filter()
def setformats(set_: ygojson.Set) -> typing.Iterable[ygojson.Format]:
    return {
        f for l in set_.contents for f in l.formats
    }  # TODO: stop using deprecated member


@app.template_filter()
def cardlocales(card: ygojson.Card) -> typing.Iterable[str]:
    return sorted(
        {
            lc
            for set_ in card.sets
            for lc in set_.locales
            if any(
                p.card.id == card.id
                for content in set_.contents
                for p in content.cards
                if set_.locales[lc] in content.locales
            )
        }
    )


@app.template_filter()
def cardformats(card: ygojson.Card) -> typing.Iterable[str]:
    return sorted(
        {
            printingformat(None, locale, content)
            for set_ in card.sets
            for content in set_.contents
            for locale in content.locales or [None]
            if any(p.card.id == card.id for p in content.cards)
        }
    )


@app.template_filter()
def cardeditions(card: ygojson.Card) -> typing.Iterable[ygojson.SetEdition]:
    return sorted(
        {
            edition
            for set_ in card.sets
            for content in set_.contents
            for edition in content.editions
            if any(p.card.id == card.id for p in content.cards)
        },
        key=lambda x: x.value,
    )


@app.template_filter()
def cardrarities(card: ygojson.Card) -> typing.Iterable[ygojson.CardRarity]:
    return sorted(
        {
            printing.rarity
            for set_ in card.sets
            for content in set_.contents
            for printing in content.cards
            if printing.card.id == card.id and printing.rarity
        },
        key=lambda x: x.value,
    )


@app.template_filter()
def printingformat(
    printing: typing.Optional[ygojson.CardPrinting],
    locale: typing.Optional[ygojson.SetLocale],
    content: ygojson.SetContents,
) -> str:
    if locale:
        return "/".join(x.value for x in locale.formats)
    return "/".join(x.value for x in content.formats)


@app.template_test()
def oftypecard(thing) -> bool:
    return type(thing) == ygojson.Card


@app.template_test()
def oftypeset(thing) -> bool:
    return type(thing) == ygojson.Set


@app.template_test()
def oftypeseries(thing) -> bool:
    return type(thing) == ygojson.Series


@app.template_test()
def oftypedistro(thing) -> bool:
    return type(thing) == ygojson.PackDistrobution


@app.template_test()
def oftypeproduct(thing) -> bool:
    return type(thing) == ygojson.SealedProduct


page_load_start_time = time.time()


@app.before_request
def before_request():
    global page_load_start_time
    page_load_start_time = time.time()


@app.after_request
def after_request(response):
    diff = time.time() - page_load_start_time
    if (
        (response.response)
        and (200 <= response.status_code < 300)
        and (response.content_type.startswith("text/html"))
    ):
        response.set_data(
            response.get_data().replace(
                b"__EXECUTION_TIME__", bytes(f"{diff:.4f}", "utf-8")
            )
        )
    return response


def common_template_vars():
    return {
        "yjv_version": __version__,
        "db_version": ygojson.__version__,
        "schema_version": ygojson.SCHEMA_VERSION,
        "accesstime": datetime.datetime.now().isoformat(),
    }


@app.route("/")
def index():
    return flask.render_template(
        "index.j2",
        **common_template_vars(),
        ygodb=ygodb,
        cards_of_the_day=cards_of_the_day,
    )


@app.route("/random-card")
def random_card():
    return app.redirect(
        flask.url_for(
            card.__name__,
            uuid=random.choice([*ygodb.cards_by_id]),
        )
    )


@app.route("/card/<uuid:uuid>")
def card(uuid: uuid.UUID):
    return flask.render_template(
        "card.j2",
        **common_template_vars(),
        ygodb=ygodb,
        card=ygodb.cards_by_id[uuid],
    )


@app.route("/random-set")
def random_set():
    return app.redirect(
        flask.url_for(
            set_.__name__,
            uuid=random.choice([*ygodb.sets_by_id]),
        )
    )


@app.route("/set/<uuid:uuid>")
def set_(uuid: uuid.UUID):
    return flask.render_template(
        "set.j2",
        **common_template_vars(),
        ygodb=ygodb,
        set=ygodb.sets_by_id[uuid],
    )


@app.route("/random-series")
def random_series():
    return app.redirect(
        flask.url_for(
            series.__name__,
            uuid=random.choice([*ygodb.series_by_id]),
        )
    )


@app.route("/series/<uuid:uuid>")
def series(uuid: uuid.UUID):
    return flask.render_template(
        "series.j2",
        **common_template_vars(),
        ygodb=ygodb,
        series=ygodb.series_by_id[uuid],
    )


@app.route("/search")
def search_():
    query = flask.request.args.get("query", "")
    page = int(flask.request.args.get("page", "1"))
    try:
        search_ = search.Search(query)
        results = search_.execute(ygodb)
        hrq = search_.human_readable_query()
        error_msg = None
    except search.SearchFailedException as e:
        results = []
        hrq = None
        error_msg = str(e)
        app.logger.exception(e)

    return flask.render_template(
        "search.j2",
        **common_template_vars(),
        ygodb=ygodb,
        query=query,
        results=results[search.SEARCH_RESULTS_PER_PAGE * (page - 1) :][
            : search.SEARCH_RESULTS_PER_PAGE
        ],
        n_results=len(results),
        SEARCH_RESULTS_PER_PAGE=search.SEARCH_RESULTS_PER_PAGE,
        page=page,
        n_pages=math.ceil(len(results) / search.SEARCH_RESULTS_PER_PAGE),
        human_readable_query=hrq,
        error_message=error_msg,
    )


@app.route("/about")
def about():
    return flask.render_template(
        "about.j2",
        **common_template_vars(),
    )


@app.route("/syntax")
def syntax():
    return flask.render_template(
        "syntax.j2",
        **common_template_vars(),
        FILTERS=search.FILTERS,
        SORTERS=search.SORTERS,
    )



import datetime
import enum
import math
import os
import sys
import typing

import lark
import ygojson

from .locales import LOCALE_TRANSLATED

SEARCH_RESULTS_PER_PAGE = 100

Thing = typing.Union[ygojson.Card, ygojson.Set, ygojson.Series, ygojson.SealedProduct]


class SearchFailedException(Exception):
    pass


class SortDir(enum.Enum):
    ASC = 0
    DESC = 1


class Sorter:
    names: typing.ClassVar[typing.List[str]]
    desc: typing.ClassVar[str]

    @classmethod
    def execute(
        cls, search: "Search", db: ygojson.Database, result: Thing, dir: SortDir
    ) -> typing.Any:
        raise NotImplementedError

    @classmethod
    def human_readable_query(cls, search: "Search") -> str:
        raise NotImplementedError


class Sort:
    sorter: typing.Type[Sorter]
    dir: SortDir

    def __init__(self, sorter: typing.Type[Sorter], dir: SortDir) -> None:
        self.sorter = sorter
        self.dir = dir

    def execute(
        self, db: ygojson.Database, search: "Search", result: Thing
    ) -> typing.Any:
        return self.sorter.execute(search, db, result, self.dir)

    def human_readable_query(self, search: "Search") -> str:
        if self.dir == SortDir.ASC:
            return self.sorter.human_readable_query(search)
        else:
            return self.sorter.human_readable_query(search) + " (descending)"


class FilterMode(enum.Enum):
    DEFAULT = ":"
    EQ = "="
    LT = "<"
    LE = "<="
    GT = ">"
    GE = ">="


class Filter:
    names: typing.ClassVar[typing.List[str]]
    desc: typing.ClassVar[str]

    @classmethod
    def execute(
        cls,
        db: ygojson.Database,
        search: "Search",
        predicate: "TermPredicate",
        results: typing.Iterable[Thing],
    ) -> typing.Iterable[Thing]:
        raise NotImplementedError

    @classmethod
    def human_readable_query(cls, search: "Search", predicate: "TermPredicate") -> str:
        raise NotImplementedError


class Term:
    def execute(
        self, db: ygojson.Database, search: "Search", results: typing.Iterable[Thing]
    ) -> typing.Iterable[Thing]:
        raise NotImplementedError

    def human_readable_query(self, search: "Search") -> str:
        raise NotImplementedError


def _flatten(xs):
    if not isinstance(xs, list):
        yield xs
        return
    for x in xs:
        yield from _flatten(x)


SORT_FILTER = "sort"
LOCALE_FILTER = "locale"


class QueryParser(lark.Transformer):
    def __init__(self, search: "Search") -> None:
        super().__init__(True)
        self.search = search

    def start(self, data) -> typing.Any:
        self.search.terms.extend(_flatten(data))

    def predicate_simple(self, data) -> typing.Any:
        (word,) = data
        return TermPredicate(FilterName, FilterMode.DEFAULT, word)

    def predicate_class(self, data) -> typing.Any:
        (cmpop, word) = data
        return TermPredicate(FilterClass, FilterMode(cmpop), word)

    def predicate_full(self, data) -> typing.Any:
        (filtername, cmpop, word) = data
        filtername_normalized = filtername.strip().lower()

        if filtername_normalized == SORT_FILTER:
            sorter_split = [x.strip() for x in word.split("-")]
            if not sorter_split:
                raise SearchFailedException(
                    f"No sorter given! You need something after the 'sort:'."
                )
            if len(sorter_split) > 2:
                raise SearchFailedException(
                    f"Too many arguments to the sorter given! The format is 'sort:SORTER[-DIRECTION]'."
                )
            sortername = sorter_split[0]
            if len(sorter_split) == 2:
                if sorter_split[1] == "asc":
                    dir_ = SortDir.ASC
                elif sorter_split[1] == "desc":
                    dir_ = SortDir.DESC
                else:
                    raise SearchFailedException(
                        f"Unknown sorting direction '{sorter_split[1]}'!"
                    )
            else:
                dir_ = SortDir.ASC
            if sortername not in SORTER_NAME_MAP:
                raise SearchFailedException(f"Unknown sorter '{filtername}'!")
            self.search.sorts.append(Sort(SORTER_NAME_MAP[sortername], dir_))
            return []

        if filtername_normalized == LOCALE_FILTER:
            self.search.locales.add(word.strip().lower())
            return []

        if filtername_normalized not in FILTER_NAME_MAP:
            raise SearchFailedException(f"Unknown filter '{filtername}'!")
        return TermPredicate(
            FILTER_NAME_MAP[filtername_normalized], FilterMode(cmpop), word
        )

    def parens(self, data) -> typing.Any:
        return data

    def alternation(self, data) -> typing.Any:
        return TermOr([*_flatten(data)])

    def negation(self, data) -> typing.Any:
        return TermNegate([*_flatten(data)])

    def WORD(self, token) -> typing.Any:
        return str(token)

    def ESCAPED_STRING(self, token) -> typing.Any:
        return str(token)[1:-1]


class Search:
    query: str
    terms: typing.List[Term]
    sorts: typing.List[Sort]
    locales: typing.Set[str]

    def __init__(self, query: str) -> None:
        self.query = query
        self.terms = []
        self.sorts = []
        self.locales = set()

        try:
            tree = LANGUAGE.parse(query)
            # print(tree.pretty())
            QueryParser(self).transform(tree)
        except lark.exceptions.VisitError as e:
            if type(e.orig_exc) is SearchFailedException:
                raise e.orig_exc
            else:
                raise
        except (lark.exceptions.LexError, lark.exceptions.ParseError) as e:
            raise SearchFailedException(f"{e}")

        if not self.sorts:
            self.sorts = [Sort(SorterClass, SortDir.ASC), Sort(SorterName, SortDir.ASC)]

    def human_readable_query(self) -> str:
        result = "all things"
        if self.terms:
            result += " "
            result += " AND ".join(x.human_readable_query(self) for x in self.terms)
        if self.sorts:
            result += ", sorted by "
            result += " and then ".join(
                x.human_readable_query(self) for x in self.sorts
            )
        if self.locales:
            result += ", in "
            result += " / ".join(LOCALE_TRANSLATED.get(x, x) for x in self.locales)
        return result

    def _exclude_cards_out_of_locale(
        self, results: typing.Iterable[ygojson.Card]
    ) -> typing.Iterable[ygojson.Card]:
        if not self.locales:
            yield from results
            return
        for result in results:
            if any(l in result.text and result.text[l].official for l in self.locales):
                yield result

    def _exclude_sets_out_of_locale(
        self, results: typing.Iterable[typing.Union[ygojson.Set, ygojson.SealedProduct]]
    ) -> typing.Iterable[typing.Union[ygojson.Set, ygojson.SealedProduct]]:
        if not self.locales:
            yield from results
            return
        for result in results:
            if any(l in result.locales for l in self.locales):
                yield result

    def _exclude_series_out_of_locale(
        self, results: typing.Iterable[ygojson.Series]
    ) -> typing.Iterable[ygojson.Series]:
        if not self.locales:
            yield from results
            return
        for result in results:
            if any(l in result.name for l in self.locales):
                yield result

    def execute(self, db: ygojson.Database) -> typing.List[Thing]:
        results = [
            *self._exclude_cards_out_of_locale(db.cards),
            *self._exclude_sets_out_of_locale(db.sets),
            *self._exclude_sets_out_of_locale(db.products),
            *self._exclude_series_out_of_locale(db.series),
        ]
        if not self.locales:
            self.locales = {"en", "jp"}
        for term in self.terms:
            results = term.execute(db, self, results)
        return sorted(
            results,
            key=lambda x: tuple(sort.execute(db, self, x) for sort in self.sorts),
        )


with open(
    os.path.join(os.path.dirname(__file__), "search.lark"), encoding="utf-8"
) as file:
    LANGUAGE = lark.Lark(file)

###################
# TERMS
###################


class TermPredicate(Term):
    filter: typing.Type[Filter]
    mode: FilterMode
    value: str

    def __init__(
        self, filter: typing.Type[Filter], mode: FilterMode, value: str
    ) -> None:
        super().__init__()
        self.filter = filter
        self.mode = mode
        self.value = value

    def execute(
        self, db: ygojson.Database, search: "Search", results: typing.Iterable[Thing]
    ) -> typing.Iterable[Thing]:
        return self.filter.execute(db, search, self, results)

    def human_readable_query(self, search: "Search") -> str:
        return self.filter.human_readable_query(search, self)


class TermOr(Term):
    terms: typing.List[Term]

    def __init__(self, terms: typing.List[Term]) -> None:
        super().__init__()
        self.terms = terms

    def execute(
        self, db: ygojson.Database, search: "Search", results: typing.Iterable[Thing]
    ) -> typing.Iterable[Thing]:
        for result in results:
            for term in self.terms:
                if [*term.execute(db, search, [result])]:
                    yield result
                    break

    def human_readable_query(self, search: "Search") -> str:
        return (
            "(" + " OR ".join(x.human_readable_query(search) for x in self.terms) + ")"
        )


class TermNegate(Term):
    terms: typing.List[Term]

    def __init__(self, terms: typing.List[Term]) -> None:
        super().__init__()
        self.terms = terms

    def execute(
        self, db: ygojson.Database, search: "Search", results: typing.Iterable[Thing]
    ) -> typing.Iterable[Thing]:
        for result in results:
            subresults = [result]
            for term in self.terms:
                subresults = [*term.execute(db, search, subresults)]
            if not subresults:
                yield result

    def human_readable_query(self, search: "Search") -> str:
        return (
            "NOT ("
            + " AND ".join(x.human_readable_query(search) for x in self.terms)
            + ")"
        )


###################
# FILTERS
###################


class FilterName(Filter):
    names = ["name", "n"]
    desc = "Filter by card name in the selected locales."

    @classmethod
    def execute(
        cls,
        db: ygojson.Database,
        search: "Search",
        predicate: "TermPredicate",
        results: typing.Iterable[Thing],
    ) -> typing.Iterable[Thing]:
        query_normalized = predicate.value.strip().lower()

        def cmp(s: str) -> bool:
            if predicate.mode == FilterMode.DEFAULT:
                return query_normalized in s
            elif predicate.mode == FilterMode.EQ:
                return query_normalized == s
            raise SearchFailedException(
                f"Search filter 'name' does not accept filter mode '{predicate.mode.value}'!"
            )

        for result in results:
            if type(result) is ygojson.Card:
                if any(
                    l in result.text and cmp(result.text[l].name.lower())
                    for l in search.locales
                ):
                    yield result
            elif type(result) is ygojson.Set:
                if any(
                    l in result.name and cmp(result.name[l].lower())
                    for l in search.locales
                ):
                    yield result
            elif type(result) is ygojson.Series:
                if any(
                    l in result.name and cmp(result.name[l].lower())
                    for l in search.locales
                ):
                    yield result
            elif type(result) is ygojson.SealedProduct:
                if any(
                    l in result.name and cmp(result.name[l].lower())
                    for l in search.locales
                ):
                    yield result

    @classmethod
    def human_readable_query(cls, search: "Search", predicate: "TermPredicate") -> str:
        if predicate.mode == FilterMode.DEFAULT:
            return f"whose name contains '{predicate.value}'"
        elif predicate.mode == FilterMode.EQ:
            return f"named '{predicate.value}'"
        else:
            return f"<ERROR: bad mode '{predicate.mode.value}'>"


class FilterEffect(Filter):
    names = ["effect", "e"]
    desc = "Filter by effect text or card lore in the selected locales."

    @classmethod
    def execute(
        cls,
        db: ygojson.Database,
        search: "Search",
        predicate: "TermPredicate",
        results: typing.Iterable[Thing],
    ) -> typing.Iterable[Thing]:
        query_normalized = predicate.value.strip().lower()

        def cmp(s: str) -> bool:
            if predicate.mode == FilterMode.DEFAULT:
                return query_normalized in s
            elif predicate.mode == FilterMode.EQ:
                return query_normalized == s
            raise SearchFailedException(
                f"Search filter 'effect' does not accept filter mode '{predicate.mode.value}'!"
            )

        for result in results:
            if type(result) is ygojson.Card:
                if any(
                    l in result.text
                    and cmp(
                        (
                            (result.text[l].pendulum_effect or "")
                            + "\n"
                            + (result.text[l].effect or "")
                        )
                        .strip()
                        .lower()
                    )
                    for l in search.locales
                ):
                    yield result

    @classmethod
    def human_readable_query(cls, search: "Search", predicate: "TermPredicate") -> str:
        if predicate.mode == FilterMode.DEFAULT:
            return f"whose effect contains '{predicate.value}'"
        elif predicate.mode == FilterMode.EQ:
            return f"whose effect is '{predicate.value}'"
        else:
            return f"<ERROR: bad mode '{predicate.mode.value}'>"


FILTER_CLASS_OPTIONS = {
    "card": ygojson.Card,
    "c": ygojson.Card,
    "set": ygojson.Set,
    "pack": ygojson.Set,
    "s": ygojson.Set,
    "product": ygojson.SealedProduct,
    "sealed": ygojson.SealedProduct,
    "sealedproduct": ygojson.SealedProduct,
    "sealed-product": ygojson.SealedProduct,
    "sealed_product": ygojson.SealedProduct,
    "p": ygojson.SealedProduct,
    "sp": ygojson.SealedProduct,
    "series": ygojson.Series,
    "archetype": ygojson.Series,
    "a": ygojson.Series,
}

THING_NAMES = {
    ygojson.Card: "cards",
    ygojson.Set: "sets",
    ygojson.SealedProduct: "sealed products",
    ygojson.Series: "series/archetypes",
}


class FilterClass(Filter):
    names = ["class", "cl", ""]
    desc = "Filter by what type of thing you want to see: <tt>card</tt>, <tt>set<tt>, <tt>product</tt>, or <tt>series</tt>."

    @classmethod
    def execute(
        cls,
        db: ygojson.Database,
        search: "Search",
        predicate: "TermPredicate",
        results: typing.Iterable[Thing],
    ) -> typing.Iterable[Thing]:
        query_normalized = predicate.value.strip().lower()
        if predicate.mode not in {FilterMode.DEFAULT, FilterMode.EQ}:
            raise SearchFailedException(
                f"Search filter 'class' does not accept filter mode '{predicate.mode.value}'!"
            )
        if query_normalized not in FILTER_CLASS_OPTIONS:
            raise SearchFailedException(
                f"""Search filter 'class' does not accept value '{query_normalized}'!
Accaptable values include 'card' (or 'c'), 'pack'/'set' or ('s'), 'sealed'/'product' (or 'p'), or 'series'/'archetype' (or 'a')."""
            )
        clazz = FILTER_CLASS_OPTIONS[query_normalized]
        for result in results:
            if type(result) is clazz:
                yield result

    @classmethod
    def human_readable_query(cls, search: "Search", predicate: "TermPredicate") -> str:
        thingtype = FILTER_CLASS_OPTIONS.get(predicate.value.strip().lower())
        if not thingtype:
            return f"<ERROR: bad value '{predicate.value}'>"
        return f"that are {THING_NAMES[thingtype]}"


class FilterType(Filter):
    names = ["type", "t"]
    desc = "Filter by the contents of a card's typeline."

    @classmethod
    def execute(
        cls,
        db: ygojson.Database,
        search: "Search",
        predicate: "TermPredicate",
        results: typing.Iterable[Thing],
    ) -> typing.Iterable[Thing]:
        query_normalized = predicate.value.strip().lower()

        def cmp(s: str) -> bool:
            if predicate.mode == FilterMode.DEFAULT:
                return query_normalized in s
            elif predicate.mode == FilterMode.EQ:
                return query_normalized in s.split("\n")
            raise SearchFailedException(
                f"Search filter 'type' does not accept filter mode '{predicate.mode.value}'!"
            )

        for result in results:
            if type(result) is ygojson.Card:
                if cmp(
                    "\n".join(
                        [
                            result.card_type.value,
                            result.type.value if result.type else "",
                            result.subcategory.value if result.subcategory else "",
                            result.character if result.character else "",
                            result.skill_type if result.skill_type else "",
                            *[t.value for t in result.monster_card_types or []],
                            *[t.value for t in result.classifications or []],
                            *[t.value for t in result.abilities or []],
                        ]
                    )
                ):
                    yield result

    @classmethod
    def human_readable_query(cls, search: "Search", predicate: "TermPredicate") -> str:
        if predicate.mode == FilterMode.DEFAULT:
            return f"whose typeline contains '{predicate.value}'"
        elif predicate.mode == FilterMode.EQ:
            return f"whose typeline contains exactly '{predicate.value}'"
        else:
            return f"<ERROR: bad mode '{predicate.mode.value}'>"


class FilterAttribute(Filter):
    names = ["attribute", "attr", "a"]
    desc = "Filter by a card's attribute."

    @classmethod
    def execute(
        cls,
        db: ygojson.Database,
        search: "Search",
        predicate: "TermPredicate",
        results: typing.Iterable[Thing],
    ) -> typing.Iterable[Thing]:
        query_normalized = predicate.value.strip().lower()

        def cmp(s: str) -> bool:
            if predicate.mode == FilterMode.DEFAULT:
                return query_normalized in s
            elif predicate.mode == FilterMode.EQ:
                return query_normalized == s
            raise SearchFailedException(
                f"Search filter 'attribute' does not accept filter mode '{predicate.mode.value}'!"
            )

        for result in results:
            if type(result) is ygojson.Card:
                if result.attribute and cmp(result.attribute.value):
                    yield result

    @classmethod
    def human_readable_query(cls, search: "Search", predicate: "TermPredicate") -> str:
        return f"whose attribute is '{predicate.value}'"


class FilterInt(Filter):
    @classmethod
    def get_int_prop(
        cls, db: ygojson.Database, predicate: "TermPredicate", result: Thing
    ) -> typing.Union[None, int, str]:
        raise NotImplementedError

    @classmethod
    def execute(
        cls,
        db: ygojson.Database,
        search: "Search",
        predicate: "TermPredicate",
        results: typing.Iterable[Thing],
    ) -> typing.Iterable[Thing]:
        query_normalized = predicate.value.strip().lower()

        def cmp(s: typing.Union[str, int]) -> bool:
            try:
                query_int = int(query_normalized)
                if (
                    predicate.mode == FilterMode.DEFAULT
                    or predicate.mode == FilterMode.EQ
                ):
                    return query_int == s
                elif predicate.mode == FilterMode.GT and type(s) is int:
                    return s > query_int
                elif predicate.mode == FilterMode.GE and type(s) is int:
                    return s >= query_int
                elif predicate.mode == FilterMode.LT and type(s) is int:
                    return s < query_int
                elif predicate.mode == FilterMode.LE and type(s) is int:
                    return s <= query_int
                return False
            except ValueError:
                if (
                    predicate.mode == FilterMode.DEFAULT
                    or predicate.mode == FilterMode.EQ
                ):
                    return query_normalized == s
                raise SearchFailedException(
                    f"Search filter '{cls.names[0]}' does not accept filter mode '{predicate.mode.value}' for non-number values!"
                )

        for result in results:
            value = cls.get_int_prop(db, predicate, result)
            if value is not None and cmp(value):
                yield result


FILTER_MODE_TO_NAME = {
    FilterMode.DEFAULT: "is",
    FilterMode.EQ: "is exactly",
    FilterMode.GT: "is greater than",
    FilterMode.LT: "is less than",
    FilterMode.GE: "is greater than or equal to",
    FilterMode.LE: "is less than or equal to",
}


class FilterATK(FilterInt):
    names = ["attack", "atk", "at"]
    desc = "Filter by cards with, greater than, or less than a certain ATK."

    @classmethod
    def get_int_prop(
        cls, db: ygojson.Database, predicate: "TermPredicate", result: Thing
    ) -> typing.Union[None, int, str]:
        if type(result) is ygojson.Card:
            return result.atk
        return None

    @classmethod
    def human_readable_query(cls, search: "Search", predicate: "TermPredicate") -> str:
        return f"whose ATK {FILTER_MODE_TO_NAME[predicate.mode]} {predicate.value}"


class FilterDEF(FilterInt):
    names = ["defence", "defense", "def", "de"]
    desc = "Filter by cards with, greater than, or less than a certain DEF."

    @classmethod
    def get_int_prop(
        cls, db: ygojson.Database, predicate: "TermPredicate", result: Thing
    ) -> typing.Union[None, int, str]:
        if type(result) is ygojson.Card:
            return result.def_
        return None

    @classmethod
    def human_readable_query(cls, search: "Search", predicate: "TermPredicate") -> str:
        return f"whose DEF {FILTER_MODE_TO_NAME[predicate.mode]} {predicate.value}"


class FilterLevel(FilterInt):
    names = ["level", "lvl", "lv", "l"]
    desc = "Filter by cards with, greater than, or less than a certain level. This does NOT match to Xyz monsters."

    @classmethod
    def get_int_prop(
        cls, db: ygojson.Database, predicate: "TermPredicate", result: Thing
    ) -> typing.Union[None, int, str]:
        if type(result) is ygojson.Card:
            return result.level
        return None

    @classmethod
    def human_readable_query(cls, search: "Search", predicate: "TermPredicate") -> str:
        return f"whose level {FILTER_MODE_TO_NAME[predicate.mode]} {predicate.value}"


class FilterRank(FilterInt):
    names = ["rank", "r"]
    desc = "Filter by cards with, greater than, or less than a certain rank. This does NOT match to non-Xyz monsters."

    @classmethod
    def get_int_prop(
        cls, db: ygojson.Database, predicate: "TermPredicate", result: Thing
    ) -> typing.Union[None, int, str]:
        if type(result) is ygojson.Card:
            return result.rank
        return None

    @classmethod
    def human_readable_query(cls, search: "Search", predicate: "TermPredicate") -> str:
        return f"whose rank {FILTER_MODE_TO_NAME[predicate.mode]} {predicate.value}"


class FilterScale(FilterInt):
    names = ["scale", "sc"]
    desc = "Filter by cards with, greater than, or less than a certain pendulum scale."

    @classmethod
    def get_int_prop(
        cls, db: ygojson.Database, predicate: "TermPredicate", result: Thing
    ) -> typing.Union[None, int, str]:
        if type(result) is ygojson.Card:
            return result.scale
        return None

    @classmethod
    def human_readable_query(cls, search: "Search", predicate: "TermPredicate") -> str:
        return f"whose pendulum scale {FILTER_MODE_TO_NAME[predicate.mode]} {predicate.value}"


class FilterLinkRating(FilterInt):
    names = ["linkrating", "link", "lr"]
    desc = "Filter by cards with, greater than, or less than a certain link rating."

    @classmethod
    def get_int_prop(
        cls, db: ygojson.Database, predicate: "TermPredicate", result: Thing
    ) -> typing.Union[None, int, str]:
        if type(result) is ygojson.Card:
            return len(result.link_arrows or [])
        return None

    @classmethod
    def human_readable_query(cls, search: "Search", predicate: "TermPredicate") -> str:
        return (
            f"whose link rating {FILTER_MODE_TO_NAME[predicate.mode]} {predicate.value}"
        )


def _date_to_timestamp(date: datetime.date) -> float:
    return datetime.datetime(date.year, date.month, date.day).timestamp()


FILTER_MODE_TO_DATE_NAME = {
    FilterMode.DEFAULT: "on",
    FilterMode.EQ: "on exactly",
    FilterMode.GT: "after",
    FilterMode.LT: "before",
    FilterMode.GE: "on or after",
    FilterMode.LE: "on or before",
}


class FilterDate(Filter):
    @classmethod
    def get_date_prop(
        cls, db: ygojson.Database, predicate: "TermPredicate", result: Thing
    ) -> typing.Optional[datetime.date]:
        raise NotImplementedError

    @classmethod
    def execute(
        cls,
        db: ygojson.Database,
        search: "Search",
        predicate: "TermPredicate",
        results: typing.Iterable[Thing],
    ) -> typing.Iterable[Thing]:
        query_normalized = predicate.value.strip().lower()

        def cmp(s: float) -> bool:
            try:
                query_timestamp = _date_to_timestamp(
                    datetime.date.fromisoformat(query_normalized)
                )
                if (
                    predicate.mode == FilterMode.DEFAULT
                    or predicate.mode == FilterMode.EQ
                ):
                    return s == query_timestamp
                elif predicate.mode == FilterMode.GT:
                    return s > query_timestamp
                elif predicate.mode == FilterMode.GE:
                    return s >= query_timestamp
                elif predicate.mode == FilterMode.LT:
                    return s < query_timestamp
                elif predicate.mode == FilterMode.LE:
                    return s <= query_timestamp
                return False
            except ValueError:
                raise SearchFailedException(
                    f"Search filter '{cls.names[0]}' does not accept non-date values!\nDates are expected in ISO format (YYYY-MM-DD)."
                )

        for result in results:
            value = cls.get_date_prop(db, predicate, result)
            if value is not None and cmp(_date_to_timestamp(value)):
                yield result


def _get_release_date(result: Thing) -> typing.Optional[datetime.date]:
    if type(result) is ygojson.Card:
        dates = []
        for set_ in result.sets:
            locales: typing.Set[ygojson.SetLocale] = set()
            for content in set_.contents:
                for printing in content.cards:
                    if printing.card == result:
                        locales.update(content.locales)
            if not locales and set_.date:
                dates.append(set_.date)
            for locale in locales:
                if locale.date:
                    dates.append(locale.date)
        if dates:
            return min(dates)
    elif type(result) is ygojson.Set or type(result) is ygojson.SealedProduct:
        if result.date:
            return result.date
        dates = [x.date for x in result.locales.values() if x.date]
        if dates:
            return min(dates)
    else:
        return None


class FilterDateOfRelease(FilterDate):
    names = ["date", "d"]
    desc = "Filter by cards and/or sets that came out for the first time at, before, or after the given date."

    @classmethod
    def get_date_prop(
        cls, db: ygojson.Database, predicate: "TermPredicate", result: Thing
    ) -> typing.Optional[datetime.date]:
        return _get_release_date(result)

    @classmethod
    def human_readable_query(cls, search: "Search", predicate: "TermPredicate") -> str:
        return f"who were released {FILTER_MODE_TO_DATE_NAME[predicate.mode]} {predicate.value}"


FILTERS = [
    FilterName,
    FilterEffect,
    FilterClass,
    FilterType,
    FilterAttribute,
    FilterATK,
    FilterDEF,
    FilterLevel,
    FilterScale,
    FilterLinkRating,
    FilterLevel,
    FilterRank,
    FilterScale,
    FilterLinkRating,
    FilterDateOfRelease,
]

FILTER_NAME_MAP = {name: filter for filter in FILTERS for name in filter.names}

###################
# SORTERS
###################


class SorterClass(Sorter):
    names = ["classes", "class", "cl"]
    desc = "Sort by what type of thing it is, in card -> set -> sealed product -> series/archetype order."

    @classmethod
    def execute(
        cls, search: "Search", db: ygojson.Database, result: Thing, dir: SortDir
    ) -> typing.Any:
        dir_factor = 1 if dir == SortDir.ASC else -1
        if type(result) is ygojson.Card:
            return 1 * dir_factor
        elif type(result) is ygojson.Set:
            return 2 * dir_factor
        elif type(result) is ygojson.Series:
            return 4 * dir_factor
        elif type(result) is ygojson.SealedProduct:
            return 3 * dir_factor
        else:
            return 5 * dir_factor

    @classmethod
    def human_readable_query(cls, search: "Search") -> str:
        return "class"


class SorterName(Sorter):
    names = ["names", "name", "n"]
    desc = "Sort by names in the selected locale."

    @classmethod
    def execute(
        cls, search: "Search", db: ygojson.Database, result: Thing, dir: SortDir
    ) -> typing.Any:
        if type(result) is ygojson.Card:
            s = "\n".join(
                (result.text[l].name.lower() if l in result.text else "�")
                for l in sorted(search.locales)
            )
        elif type(result) is ygojson.Set:
            s = "\n".join(
                (result.name[l].lower() if l in result.name else "�")
                for l in sorted(search.locales)
            )
        elif type(result) is ygojson.Series:
            s = "\n".join(
                (result.name[l].lower() if l in result.name else "�")
                for l in sorted(search.locales)
            )
        elif type(result) is ygojson.SealedProduct:
            s = "\n".join(
                (result.name[l].lower() if l in result.name else "�")
                for l in sorted(search.locales)
            )
        else:
            return None

        if dir == SortDir.ASC:
            return s
        else:
            return tuple(-ord(c) for c in s)

    @classmethod
    def human_readable_query(cls, search: "Search") -> str:
        return "name"


class SorterATK(Sorter):
    names = ["attack", "atk", "at"]
    desc = "Sort by ATK."

    @classmethod
    def execute(
        cls, search: "Search", db: ygojson.Database, result: Thing, dir: SortDir
    ) -> typing.Any:
        if type(result) is ygojson.Card:
            v = result.atk
            if v is None:
                return sys.maxsize
            elif type(v) is str:
                return sys.maxsize - 1
            elif dir == SortDir.ASC:
                return v
            else:
                return -int(v)
        else:
            return sys.maxsize

    @classmethod
    def human_readable_query(cls, search: "Search") -> str:
        return "ATK"


class SorterDEF(Sorter):
    names = ["defence", "defense", "def", "de"]
    desc = "Sort by DEF."

    @classmethod
    def execute(
        cls, search: "Search", db: ygojson.Database, result: Thing, dir: SortDir
    ) -> typing.Any:
        if type(result) is ygojson.Card:
            v = result.def_
            if v is None:
                return sys.maxsize
            elif type(v) is str:
                return sys.maxsize - 1
            elif dir == SortDir.ASC:
                return v
            else:
                return -int(v)
        else:
            return sys.maxsize

    @classmethod
    def human_readable_query(cls, search: "Search") -> str:
        return "DEF"


class SorterLevel(Sorter):
    names = ["level", "lvl", "lv", "l"]
    desc = "Sort by level. Does NOT sort ranks in with levels."

    @classmethod
    def execute(
        cls, search: "Search", db: ygojson.Database, result: Thing, dir: SortDir
    ) -> typing.Any:
        if type(result) is ygojson.Card:
            v = result.level
            if v is None:
                return sys.maxsize
            elif dir == SortDir.ASC:
                return v
            else:
                return -int(v)
        else:
            return sys.maxsize

    @classmethod
    def human_readable_query(cls, search: "Search") -> str:
        return "level"


class SorterRank(Sorter):
    names = ["rank", "r"]
    desc = "Sort by rank. Does NOT sort levels in with ranks."

    @classmethod
    def execute(
        cls, search: "Search", db: ygojson.Database, result: Thing, dir: SortDir
    ) -> typing.Any:
        if type(result) is ygojson.Card:
            v = result.rank
            if v is None:
                return sys.maxsize
            elif dir == SortDir.ASC:
                return v
            else:
                return -int(v)
        else:
            return sys.maxsize

    @classmethod
    def human_readable_query(cls, search: "Search") -> str:
        return "rank"


class SorterScale(Sorter):
    names = ["scale", "sc"]
    desc = "Sort by pendulum scale."

    @classmethod
    def execute(
        cls, search: "Search", db: ygojson.Database, result: Thing, dir: SortDir
    ) -> typing.Any:
        if type(result) is ygojson.Card:
            v = result.scale
            if v is None:
                return sys.maxsize
            elif dir == SortDir.ASC:
                return v
            else:
                return -int(v)
        else:
            return sys.maxsize

    @classmethod
    def human_readable_query(cls, search: "Search") -> str:
        return "pendulum scale"


class SorterLink(Sorter):
    names = ["linkranking", "link", "lr"]
    desc = "Sort by link rating."

    @classmethod
    def execute(
        cls, search: "Search", db: ygojson.Database, result: Thing, dir: SortDir
    ) -> typing.Any:
        if type(result) is ygojson.Card:
            v = len(result.link_arrows or [])
            if dir == SortDir.ASC:
                return v
            else:
                return -int(v)
        else:
            return sys.maxsize

    @classmethod
    def human_readable_query(cls, search: "Search") -> str:
        return "link rating"


class SorterDate(Sorter):
    names = ["date", "d"]
    desc = "Sort by first release date."

    @classmethod
    def execute(
        cls, search: "Search", db: ygojson.Database, result: Thing, dir: SortDir
    ) -> typing.Any:
        date = _get_release_date(result)
        if date is None:
            return math.inf
        elif dir == SortDir.ASC:
            return _date_to_timestamp(date)
        elif dir == SortDir.DESC:
            return -_date_to_timestamp(date)

    @classmethod
    def human_readable_query(cls, search: "Search") -> str:
        return "release date"


SORTERS = [
    SorterClass,
    SorterName,
    SorterATK,
    SorterDEF,
    SorterLevel,
    SorterRank,
    SorterScale,
    SorterLink,
    SorterDate,
]

SORTER_NAME_MAP = {name: sorter for sorter in SORTERS for name in sorter.names}



import os
import pathlib

import setuptools
import setuptools.command.build_py
import setuptools.command.install
import setuptools.command.sdist

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")

PKG_NAME = "yjviewer"

exec((here / "src" / PKG_NAME / "version.py").read_text(encoding="utf-8"))

setuptools.setup(
    name=PKG_NAME,
    version=__version__,
    description="A web application that lets you view YGOJSON data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/iconmaster5326/YJViewer",
    author="iconmaster5326",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: End Users/Desktop",
        "Topic :: Games/Entertainment",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    keywords="yugioh,ygo,ygojson",
    packages=setuptools.find_packages("src"),
    python_requires=">=3.8, <4",
    install_requires=["Flask", "lark", "ygojson"],
    extras_require={
        "dev": ["pre-commit", "watchdog"],
        "test": [],
    },
    package_dir={
        "": "src",
    },
    # entry_points={
    #     "console_scripts": [
    #         f"{PKG_NAME}={PKG_NAME}.__main__:main",
    #     ],
    # },
)



LOCALE_TRANSLATED = {
    "en": "English",
    "it": "Italian",
    "jp": "Japanese",
    "sc": "Simplified Chinese",
    "tc": "Traditional Chinese",
    "ae": "Asian-English",
    "eu": "English (Europe)",
    "na": "English (North America)",
    "fr": "French",
    "fc": "French (Canada)",
    "sp": "Spanish",
    "de": "German",
    "pt": "Portugese",
    "ja": "Japanese",
    "au": "English (Austrailia)",
    "kr": "Korean",
}
